#include "../mdk/mdk.h"
#include <string.h>
#include <windows.h>
#include "resource.h"

#pragma optimize ("awy", on) 

CMachineParameter const paraVelocity =
{
	pt_byte,
	"Vel. Sens.",
	"Velocity Sensivity",
	0,
	0xFE,
	0xFF,
	MPF_STATE,
	0x7F
};

CMachineParameter const paraType =
{
	pt_byte,
	"SustainType",
	"Sustain Type",
	0,
	1,
	0xFF,
	MPF_STATE,
	0
};

CMachineParameter const paraTrkNote =
{
	pt_note,
	"Note",
	"Note",
	NOTE_MIN,
	NOTE_OFF,
	NOTE_NO,
	0,
	0
};

CMachineParameter const paraTrkVel =
{
	pt_byte,
	"Velocity",
	"Velocity",
	0,
	0x7F,
	0xFF,
	0,
	0
};

CMachineParameter const paraTrkNoteDelay = 
{
	pt_byte,
	"Note Delay",
	"Note Delay",
	0,
	0xFE,
	0xFF,
	0,
	0
};

CMachineAttribute const attrMIDIChannel = {
	"MIDI Channel",
	0,
	16,
	0
};

//CMachineAttribute const attrMIDIVelocity = {
//	"MIDI Velocity",
//	0,
//	5,
//	0
//};

CMachineAttribute const attrMIDITranspose = {
	"MIDI Transpose (-24..24 smt)",
	0,
	48,
	24
};

CMachineAttribute const attrMIDIRecordable = {
	"MIDI Recording Enabled",
	0,
	1,
	1
};

CMachineAttribute const attrMIDINoteDelayUsed = {
	"MIDI Note Delay Used",
	0,
	1,
	1
};

CMachineAttribute const attrMode = {
	"Mode (0=note type, 1=trigger)",
	0,
	1,
	0
};
CMachineAttribute const attrPreferedTrack = {
	"Prefered Track (0=none)",
	0,
	12,
	0
};

CMachineParameter const *pParameters[] = {
	&paraVelocity, &paraType,
	&paraTrkNote, &paraTrkVel, &paraTrkNoteDelay
};
CMachineAttribute const *pAttributes[] = {
	&attrMIDIChannel,
	&attrMIDITranspose,
	&attrMIDIRecordable,
	&attrMIDINoteDelayUsed,

	&attrMode,
	&attrPreferedTrack
};

#pragma pack(1)		

class gvals
{
public:
	byte veltype;
	byte type;
};

class tvals
{
public:
	byte note;
	byte velo;
	byte notedelay;
};
class avals
{
public:
	int midichannel;
	int miditranspose;
	int midirecord;
	int notedelayused;

	int triggermode;
	int prefertrack;
};

#pragma pack()


CMachineInfo const MacInfo = 
{
	MT_EFFECT,
	MI_VERSION,	
	MIF_DOES_INPUT_MIXING | MIF_NO_OUTPUT | MIF_CONTROL_MACHINE,
	12,										// min tracks
	12,										// max tracks
	2,										// numGlobalParameters
	3,										// numTrackParameters
	pParameters,
	6,
	pAttributes,
	"CyanPhase mC-1",					// name
	"mC-1",							// short name
	"Edward L. Blake",						// author
	"&How to use...\n&Edit velocity Name...\n&About..."
};

class mi;

class miex : public CMDKMachineInterfaceEx 
{
public:
	virtual void AddInput(char const *macname, bool stereo);	// called when input is added to a machine
	virtual void DeleteInput(char const *macename);			
	virtual void RenameInput(char const *macoldname, char const *macnewname); 
	virtual void SetInputChannels(char const *macname, bool stereo); //{}
	virtual void Input(float *psamples, int numsamples, float amp) {} // if MIX_DOES_INPUT_MIXING
	virtual bool HandleInput(int index, int amp, int pan); //{ return false; }
public:

	mi *pmi;
};

class CTrack {
public:
	int NoteOn;
	int my_velo;
	byte CurrentNote;
};

class mi : public CMDKMachineInterface
{
public:
	mi();
	virtual ~mi();
	virtual void Tick();
	virtual void SetNumTracks(int const n) { };
	virtual void MDKInit(CMachineDataInput * const pi);
	virtual bool MDKWork(float *psamples, int numsamples, int const mode);
	virtual bool MDKWorkStereo(float *psamples, int numsamples, int const mode);
	virtual void Command(int const i);
	virtual void MDKSave(CMachineDataOutput * const po);
	virtual char const *DescribeValue(int const param, int const value);

public:
	virtual CMDKMachineInterfaceEx *GetEx() { return &ex; }
	virtual void OutputModeChanged(bool stereo) {}

	virtual int AllocateTrack(CSequence *pseq, int note);
	virtual void MidiNoteOff(int c, CSequence *pseq, int notedelay);
	virtual void MidiNote(int const channel, int const value, int const velocity);

public:
	miex ex;

	char machine_name1[64];
	int  machine_used1;
	CMachine *machine_ref1;
	char machine_name2[64];
	int  machine_used2;
	CMachine *machine_ref2;
	char machine_name3[64];
	int  machine_used3;
	CMachine *machine_ref3;
	char machine_name4[64];
	int  machine_used4;
	CMachine *machine_ref4;
	char machine_name5[64];
	int  machine_used5;
	CMachine *machine_ref5;
	char machine_name6[64];
	int  machine_used6;
	CMachine *machine_ref6;
	char machine_name7[64];
	int  machine_used7;
	CMachine *machine_ref7;
	char machine_name8[64];
	int  machine_used8;
	CMachine *machine_ref8;
	char machine_name9[64];
	int  machine_used9;
	CMachine *machine_ref9;
	char machine_name10[64];
	int  machine_used10;
	CMachine *machine_ref10;
	char machine_name11[64];
	int  machine_used11;
	CMachine *machine_ref11;
	char machine_name12[64];
	int  machine_used12;
	CMachine *machine_ref12;

	char delay_name1[64];
	int  delay_used1;
	CMachine *delay_ref1;
	char delay_name2[64];
	int  delay_used2;
	CMachine *delay_ref2;
	char delay_name3[64];
	int  delay_used3;
	CMachine *delay_ref3;
	char delay_name4[64];
	int  delay_used4;
	CMachine *delay_ref4;
	char delay_name5[64];
	int  delay_used5;
	CMachine *delay_ref5;
	char delay_name6[64];
	int  delay_used6;
	CMachine *delay_ref6;
	char delay_name7[64];
	int  delay_used7;
	CMachine *delay_ref7;
	char delay_name8[64];
	int  delay_used8;
	CMachine *delay_ref8;
	char delay_name9[64];
	int  delay_used9;
	CMachine *delay_ref9;
	char delay_name10[64];
	int  delay_used10;
	CMachine *delay_ref10;
	char delay_name11[64];
	int  delay_used11;
	CMachine *delay_ref11;
	char delay_name12[64];
	int  delay_used12;
	CMachine *delay_ref12;

public:
	int sustaintype;
	float velocitysens;
	char vel_name[80];

public:
	CTrack Tracks[12];
	gvals gval;
	avals aval;
	tvals tval[12];
	CMachine *ThisMachine;

};



mi::mi() {  
	GlobalVals = &gval;	
	TrackVals = tval; 
	AttrVals = (int *)&aval;
}
mi::~mi() { }

void mi::MDKInit(CMachineDataInput * const pi)
{
	int my_ver, i;
	SetOutputMode( true );	//	If true, the MDKWork will never be called, meaning that Buzz will convert a mono signal to
							//	stereo itself and call MDKWorkStereo insted.
							//	If false, MDKWork will be called in mono cases, and the output should be mono
	sprintf(machine_name1,""); sprintf(machine_name2,"");
	sprintf(machine_name3,""); sprintf(machine_name4,"");
	sprintf(machine_name5,""); sprintf(machine_name6,"");
	sprintf(machine_name7,""); sprintf(machine_name8,"");
	sprintf(machine_name9,""); sprintf(machine_name10,"");
	sprintf(machine_name11,""); sprintf(machine_name12,"");
	sprintf(delay_name1, ""); sprintf(delay_name2, "");
	sprintf(delay_name3, ""); sprintf(delay_name4, "");
	sprintf(delay_name5, ""); sprintf(delay_name6, "");
	sprintf(delay_name7, ""); sprintf(delay_name8, "");
	sprintf(delay_name9, ""); sprintf(delay_name10, "");
	sprintf(delay_name11, ""); sprintf(delay_name12, "");
	velocitysens = 1.0f;
	sprintf(vel_name, "Volume");

	machine_used1 = 0; machine_used2 = 0;
	machine_used3 = 0; machine_used4 = 0;
	machine_used5 = 0; machine_used6 = 0;
	machine_used7 = 0; machine_used8 = 0;
	machine_used9 = 0; machine_used10 = 0;
	machine_used11 = 0; machine_used12 = 0;
	delay_used1 = 0; delay_used2 = 0;
	delay_used3 = 0; delay_used4 = 0;
	delay_used5 = 0; delay_used6 = 0;
	delay_used7 = 0; delay_used8 = 0;
	delay_used9 = 0; delay_used10 = 0;
	delay_used11 = 0; delay_used12 = 0;

	sustaintype = 0;

	for (i = 0; i < 12; i++) {
		Tracks[i].CurrentNote = 0;
//		Tracks[i].CurrentDelayNote = 0;

		Tracks[i].NoteOn = 0;
		Tracks[i].my_velo = 127;
//		Tracks[i].currentsample = 0;
//		Tracks[i].isondelay = 0;
//		Tracks[i].targetsample = 0;
	};

	if (pi != NULL) {
		pi->Read(my_ver);

		if (my_ver == 100) {
			pi->Read(vel_name, 80);
		};
	};

	ThisMachine = pCB->GetThisMachine();

	ex.pmi = this;

}

void miex::AddInput(char const *macname, bool stereo) {
	if (macname != NULL) {
		if (strcmp(pmi->pCB->GetMachineInfo(pmi->pCB->GetMachine(macname))->Name, "CyanPhase mC-Delay") == 0) {
			if (pmi->delay_used1 == 0) { sprintf(pmi->delay_name1, "%s", macname); pmi->delay_used1 = 1; pmi->delay_ref1 = pmi->pCB->GetMachine(pmi->delay_name1); return; };
			if (pmi->delay_used2 == 0) { sprintf(pmi->delay_name2, "%s", macname); pmi->delay_used2 = 1; pmi->delay_ref2 = pmi->pCB->GetMachine(pmi->delay_name2); return; };
			if (pmi->delay_used3 == 0) { sprintf(pmi->delay_name3, "%s", macname); pmi->delay_used3 = 1; pmi->delay_ref3 = pmi->pCB->GetMachine(pmi->delay_name3); return; };
			if (pmi->delay_used4 == 0) { sprintf(pmi->delay_name4, "%s", macname); pmi->delay_used4 = 1; pmi->delay_ref4 = pmi->pCB->GetMachine(pmi->delay_name4); return; };
			if (pmi->delay_used5 == 0) { sprintf(pmi->delay_name5, "%s", macname); pmi->delay_used5 = 1; pmi->delay_ref5 = pmi->pCB->GetMachine(pmi->delay_name5); return; };
			if (pmi->delay_used6 == 0) { sprintf(pmi->delay_name6, "%s", macname); pmi->delay_used6 = 1; pmi->delay_ref6 = pmi->pCB->GetMachine(pmi->delay_name6); return; };
			if (pmi->delay_used7 == 0) { sprintf(pmi->delay_name7, "%s", macname); pmi->delay_used7 = 1; pmi->delay_ref7 = pmi->pCB->GetMachine(pmi->delay_name7); return; };
			if (pmi->delay_used8 == 0) { sprintf(pmi->delay_name8, "%s", macname); pmi->delay_used8 = 1; pmi->delay_ref8 = pmi->pCB->GetMachine(pmi->delay_name8); return; };
			if (pmi->delay_used9 == 0) { sprintf(pmi->delay_name9, "%s", macname); pmi->delay_used9 = 1; pmi->delay_ref9 = pmi->pCB->GetMachine(pmi->delay_name9); return; };
			if (pmi->delay_used10 == 0) { sprintf(pmi->delay_name10, "%s", macname); pmi->delay_used10 = 1; pmi->delay_ref10 = pmi->pCB->GetMachine(pmi->delay_name10); return; };
			if (pmi->delay_used11 == 0) { sprintf(pmi->delay_name11, "%s", macname); pmi->delay_used11 = 1; pmi->delay_ref11 = pmi->pCB->GetMachine(pmi->delay_name11); return; };
			if (pmi->delay_used12 == 0) { sprintf(pmi->delay_name12, "%s", macname); pmi->delay_used12 = 1; pmi->delay_ref12 = pmi->pCB->GetMachine(pmi->delay_name12); return; };
		} else {
			if (pmi->machine_used1 == 0) { sprintf(pmi->machine_name1, "%s", macname); pmi->machine_used1 = 1; pmi->machine_ref1 = pmi->pCB->GetMachine(pmi->machine_name1); return; };
			if (pmi->machine_used2 == 0) { sprintf(pmi->machine_name2, "%s", macname); pmi->machine_used2 = 1; pmi->machine_ref2 = pmi->pCB->GetMachine(pmi->machine_name2); return; };
			if (pmi->machine_used3 == 0) { sprintf(pmi->machine_name3, "%s", macname); pmi->machine_used3 = 1; pmi->machine_ref3 = pmi->pCB->GetMachine(pmi->machine_name3); return; };
			if (pmi->machine_used4 == 0) { sprintf(pmi->machine_name4, "%s", macname); pmi->machine_used4 = 1; pmi->machine_ref4 = pmi->pCB->GetMachine(pmi->machine_name4); return; };
			if (pmi->machine_used5 == 0) { sprintf(pmi->machine_name5, "%s", macname); pmi->machine_used5 = 1; pmi->machine_ref5 = pmi->pCB->GetMachine(pmi->machine_name5); return; };
			if (pmi->machine_used6 == 0) { sprintf(pmi->machine_name6, "%s", macname); pmi->machine_used6 = 1; pmi->machine_ref6 = pmi->pCB->GetMachine(pmi->machine_name6); return; };
			if (pmi->machine_used7 == 0) { sprintf(pmi->machine_name7, "%s", macname); pmi->machine_used7 = 1; pmi->machine_ref7 = pmi->pCB->GetMachine(pmi->machine_name7); return; };
			if (pmi->machine_used8 == 0) { sprintf(pmi->machine_name8, "%s", macname); pmi->machine_used8 = 1; pmi->machine_ref8 = pmi->pCB->GetMachine(pmi->machine_name8); return; };
			if (pmi->machine_used9 == 0) { sprintf(pmi->machine_name9, "%s", macname); pmi->machine_used9 = 1; pmi->machine_ref9 = pmi->pCB->GetMachine(pmi->machine_name9); return; };
			if (pmi->machine_used10 == 0) { sprintf(pmi->machine_name10, "%s", macname); pmi->machine_used10 = 1; pmi->machine_ref10 = pmi->pCB->GetMachine(pmi->machine_name10); return; };
			if (pmi->machine_used11 == 0) { sprintf(pmi->machine_name11, "%s", macname); pmi->machine_used11 = 1; pmi->machine_ref11 = pmi->pCB->GetMachine(pmi->machine_name11); return; };
			if (pmi->machine_used12 == 0) { sprintf(pmi->machine_name12, "%s", macname); pmi->machine_used12 = 1; pmi->machine_ref12 = pmi->pCB->GetMachine(pmi->machine_name12); return; };
		};
	};
}

void miex::DeleteInput(char const *macename) {
	if (macename != NULL) {
		if (strcmp(pmi->delay_name1,macename) == 0) { sprintf(pmi->delay_name1,""); pmi->delay_used1 = 0; return; };
		if (strcmp(pmi->delay_name2,macename) == 0) { sprintf(pmi->delay_name2,""); pmi->delay_used2 = 0; return; };
		if (strcmp(pmi->delay_name3,macename) == 0) { sprintf(pmi->delay_name3,""); pmi->delay_used3 = 0; return; };
		if (strcmp(pmi->delay_name4,macename) == 0) { sprintf(pmi->delay_name4,""); pmi->delay_used4 = 0; return; };
		if (strcmp(pmi->delay_name5,macename) == 0) { sprintf(pmi->delay_name5,""); pmi->delay_used5 = 0; return; };
		if (strcmp(pmi->delay_name6,macename) == 0) { sprintf(pmi->delay_name6,""); pmi->delay_used6 = 0; return; };
		if (strcmp(pmi->delay_name7,macename) == 0) { sprintf(pmi->delay_name7,""); pmi->delay_used7 = 0; return; };
		if (strcmp(pmi->delay_name8,macename) == 0) { sprintf(pmi->delay_name8,""); pmi->delay_used8 = 0; return; };
		if (strcmp(pmi->delay_name9,macename) == 0) { sprintf(pmi->delay_name9,""); pmi->delay_used9 = 0; return; };
		if (strcmp(pmi->delay_name10,macename) == 0) { sprintf(pmi->delay_name10,""); pmi->delay_used10 = 0; return; };
		if (strcmp(pmi->delay_name11,macename) == 0) { sprintf(pmi->delay_name11,""); pmi->delay_used11 = 0; return; };
		if (strcmp(pmi->delay_name12,macename) == 0) { sprintf(pmi->delay_name12,""); pmi->delay_used12 = 0; return; };

		if (strcmp(pmi->machine_name1,macename) == 0) { sprintf(pmi->machine_name1,""); pmi->machine_used1 = 0; return; };
		if (strcmp(pmi->machine_name2,macename) == 0) { sprintf(pmi->machine_name2,""); pmi->machine_used2 = 0; return; };
		if (strcmp(pmi->machine_name3,macename) == 0) { sprintf(pmi->machine_name3,""); pmi->machine_used3 = 0; return; };
		if (strcmp(pmi->machine_name4,macename) == 0) { sprintf(pmi->machine_name4,""); pmi->machine_used4 = 0; return; };
		if (strcmp(pmi->machine_name5,macename) == 0) { sprintf(pmi->machine_name5,""); pmi->machine_used5 = 0; return; };
		if (strcmp(pmi->machine_name6,macename) == 0) { sprintf(pmi->machine_name6,""); pmi->machine_used6 = 0; return; };
		if (strcmp(pmi->machine_name7,macename) == 0) { sprintf(pmi->machine_name7,""); pmi->machine_used7 = 0; return; };
		if (strcmp(pmi->machine_name8,macename) == 0) { sprintf(pmi->machine_name8,""); pmi->machine_used8 = 0; return; };
		if (strcmp(pmi->machine_name9,macename) == 0) { sprintf(pmi->machine_name9,""); pmi->machine_used9 = 0; return; };
		if (strcmp(pmi->machine_name10,macename) == 0) { sprintf(pmi->machine_name10,""); pmi->machine_used10 = 0; return; };
		if (strcmp(pmi->machine_name11,macename) == 0) { sprintf(pmi->machine_name11,""); pmi->machine_used11 = 0; return; };
		if (strcmp(pmi->machine_name12,macename) == 0) { sprintf(pmi->machine_name12,""); pmi->machine_used12 = 0; return; };
	};
}			

void miex::RenameInput(char const *macoldname, char const *macnewname) {
	if (macoldname != NULL) {
		if (strcmp(pmi->delay_name1,macoldname) == 0) { sprintf(pmi->delay_name1,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name2,macoldname) == 0) { sprintf(pmi->delay_name2,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name3,macoldname) == 0) { sprintf(pmi->delay_name3,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name4,macoldname) == 0) { sprintf(pmi->delay_name4,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name5,macoldname) == 0) { sprintf(pmi->delay_name5,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name6,macoldname) == 0) { sprintf(pmi->delay_name6,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name7,macoldname) == 0) { sprintf(pmi->delay_name7,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name8,macoldname) == 0) { sprintf(pmi->delay_name8,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name9,macoldname) == 0) { sprintf(pmi->delay_name9,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name10,macoldname) == 0) { sprintf(pmi->delay_name10,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name11,macoldname) == 0) { sprintf(pmi->delay_name11,"%s",macnewname); return; };
		if (strcmp(pmi->delay_name12,macoldname) == 0) { sprintf(pmi->delay_name12,"%s",macnewname); return; };

		if (strcmp(pmi->machine_name1,macoldname) == 0) { sprintf(pmi->machine_name1,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name2,macoldname) == 0) { sprintf(pmi->machine_name2,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name3,macoldname) == 0) { sprintf(pmi->machine_name3,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name4,macoldname) == 0) { sprintf(pmi->machine_name4,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name5,macoldname) == 0) { sprintf(pmi->machine_name5,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name6,macoldname) == 0) { sprintf(pmi->machine_name6,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name7,macoldname) == 0) { sprintf(pmi->machine_name7,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name8,macoldname) == 0) { sprintf(pmi->machine_name8,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name9,macoldname) == 0) { sprintf(pmi->machine_name9,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name10,macoldname) == 0) { sprintf(pmi->machine_name10,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name11,macoldname) == 0) { sprintf(pmi->machine_name11,"%s",macnewname); return; };
		if (strcmp(pmi->machine_name12,macoldname) == 0) { sprintf(pmi->machine_name12,"%s",macnewname); return; };
	};
} 

void miex::SetInputChannels(char const *macname, bool stereo) { }
bool miex::HandleInput(int index, int amp, int pan) { return true; }
void mi::MDKSave(CMachineDataOutput * const po) { 
	int my_ver = 100;
	po->Write(my_ver);
	po->Write(vel_name,80);
};

void mi::Tick() {
	int i, i2, u, u2, u3, u4, glb;

	if (gval.type != 0xFF) { sustaintype = gval.type; };
	if (gval.veltype != 0xFF) {
		velocitysens = gval.veltype;
	};

	for (i2 = 0; i2 < 12; i2++) {
		if (tval[i2].velo != 0xFF) Tracks[i2].my_velo = tval[i2].velo;

		if (tval[i2].notedelay != 0xFF) {
			if (delay_used1 == 1) { pCB->ControlChange(delay_ref1, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used2 == 1) { pCB->ControlChange(delay_ref2, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used3 == 1) { pCB->ControlChange(delay_ref3, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used4 == 1) { pCB->ControlChange(delay_ref4, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used5 == 1) { pCB->ControlChange(delay_ref5, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used6 == 1) { pCB->ControlChange(delay_ref6, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used7 == 1) { pCB->ControlChange(delay_ref7, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used8 == 1) { pCB->ControlChange(delay_ref8, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used9 == 1) { pCB->ControlChange(delay_ref9, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used10 == 1) { pCB->ControlChange(delay_ref10, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used11 == 1) { pCB->ControlChange(delay_ref11, 1, 0, 1 + i2, tval[i2].notedelay); };
			if (delay_used12 == 1) { pCB->ControlChange(delay_ref12, 1, 0, 1 + i2, tval[i2].notedelay); };
		};

		if (tval[i2].note != NOTE_NO) {
			if (machine_used1 == 1) {
				glb = pCB->GetMachineInfo(machine_ref1)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref1)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref1)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref1, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref1, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref1)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref1)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref1)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref1, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref1, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used2 == 1) {
				glb = pCB->GetMachineInfo(machine_ref2)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref2)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref2)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref2, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref2, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref2)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref2)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref2)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref2, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref2, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used3 == 1) {
				glb = pCB->GetMachineInfo(machine_ref3)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref3)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref3)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref3, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref3, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref3)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref3)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref3)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref3, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref3, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used4 == 1) {
				glb = pCB->GetMachineInfo(machine_ref4)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref4)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref4)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref4, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref4, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref4)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref4)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref4)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref4, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref4, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used5 == 1) {
				glb = pCB->GetMachineInfo(machine_ref5)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref5)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref5)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref5, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref5, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref5)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref5)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref5)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref5, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref5, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used6 == 1) {
				glb = pCB->GetMachineInfo(machine_ref6)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref6)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref6)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref6, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref6, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref6)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref6)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref6)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref6, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref6, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used7 == 1) {
				glb = pCB->GetMachineInfo(machine_ref7)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref7)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref7)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref7, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref7, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref7)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref7)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref7)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref7, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref7, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used8 == 1) {
				glb = pCB->GetMachineInfo(machine_ref8)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref8)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref8)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref8, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref8, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref8)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref8)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref8)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref8, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref8, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used9 == 1) {
				glb = pCB->GetMachineInfo(machine_ref9)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref9)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref9)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref9, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref9, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref9)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref9)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref9)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref9, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref9, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used10 == 1) {
				glb = pCB->GetMachineInfo(machine_ref10)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref10)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref10)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref10, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref10, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref10)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref10)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref10)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref10, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref10, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used11 == 1) {
				glb = pCB->GetMachineInfo(machine_ref11)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref11)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref11)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref11, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref11, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref11)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref11)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref11)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref11, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref11, 1, 0, i, u3);
						}
					}
				}
			};
			if (machine_used12 == 1) {
				glb = pCB->GetMachineInfo(machine_ref12)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref12)->numTrackParameters);
				for (i = 0; i < u; i++) {
					if (pCB->GetMachineInfo(machine_ref12)->Parameters[i]->Type == pt_note) {
						if (i >= glb) {	
							pCB->ControlChange(machine_ref12, 2, i2, i - glb, tval[i2].note);
						} else {
							pCB->ControlChange(machine_ref12, 1, 0, i, tval[i2].note);
						}
					}
				}
				for (i = 0; i < u; i++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref12)->Parameters[i]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref12)->Parameters[i]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)Tracks[i2].my_velo * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref12)->Parameters[i]->MaxValue - u2) / 127 + u2;
						if (i >= glb) {	
							pCB->ControlChange(machine_ref12, 2, i2, i - glb, u3);
						} else {
							pCB->ControlChange(machine_ref12, 1, 0, i, u3);
						}
					}
				}
			};
		};
	};
}

bool mi::MDKWork(float *psamples, int numsamples, int const mode) { return false; }
bool mi::MDKWorkStereo(float *psamples, int numsamples, int const mode) { return false; }

HINSTANCE dllInstance;
mi *g_mi;

BOOL WINAPI DllMain ( HANDLE hModule, DWORD fwdreason, LPVOID lpReserved )
{
	switch (fwdreason) {
	case DLL_PROCESS_ATTACH: dllInstance = (HINSTANCE) hModule; break;
	case DLL_THREAD_ATTACH: break;
	case DLL_THREAD_DETACH: break;
	case DLL_PROCESS_DETACH: break;
	}
	return TRUE;
}

BOOL APIENTRY GeneralDialog(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
	case WM_INITDIALOG: return 1;
	case WM_SHOWWINDOW: return 1;
	case WM_CLOSE: EndDialog (hDlg, TRUE);
	case WM_COMMAND:
		switch ( LOWORD (wParam))
		{
		case IDOK:
			EndDialog(hDlg, TRUE);
			return 1;
		default:
			return 0;
		}
		break;
	}
	return 0;
}

BOOL APIENTRY EditVelDialog(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
	case WM_INITDIALOG:
		SetDlgItemText(hDlg, IDC_VELNAME, g_mi->vel_name);
		return 1;
	case WM_SHOWWINDOW: return 1;
	case WM_CLOSE:
		GetDlgItemText(hDlg, IDC_VELNAME, g_mi->vel_name, 80);
		EndDialog (hDlg, TRUE);
	case WM_COMMAND:
		switch ( LOWORD (wParam))
		{
		case IDOK:
			GetDlgItemText(hDlg, IDC_VELNAME, g_mi->vel_name, 80);
			EndDialog(hDlg, TRUE);
			return 1;
		case IDCANCEL:
			EndDialog(hDlg, TRUE);
			return 1;
		default:
			return 0;
		}
		break;
	}
	return 0;
}

void mi::Command(int const i)
{
	switch (i)
	{
	case 0:
		g_mi=this;
		DialogBox(dllInstance, MAKEINTRESOURCE (IDD_HOWTO), GetForegroundWindow(), (DLGPROC) &GeneralDialog);
		break;
	case 1:
		g_mi=this;
		DialogBox(dllInstance, MAKEINTRESOURCE (IDD_EDITVELNAME), GetForegroundWindow(), (DLGPROC) &EditVelDialog);
		break;
	case 2:
		MessageBox(NULL,"CyanPhase mC-1 1.0\n\nCopyright 2001 Edward L. Blake\nEmail: blakee@rovoscape.com\n\nUses the Utitity Skin 2001 skin","About CyanPhase mC-1",MB_OK|MB_SYSTEMMODAL);
		break;
	default:
		break;
	}
}
char const *mi::DescribeValue(int const param, int const value)
{
	static char txt[16];
	switch(param)
	{
	case 0:
		sprintf(txt, "%.2f%%", (float)(value / 2.54f));
		return txt;
		break;
	case 1:
		switch (value) {
		case 0: return "Sustain"; break;
		case 1: return "Non-Sustain"; break;
		default: return NULL; break;
		}
		break;
	case 2: return NULL;
	case 3: 
		sprintf(txt, "%.2f%%", (float)(value / 2.54f));
		return txt;
		break;

	case 4:
		sprintf(txt, "%.2f ticks (%i/255th of tick)", (float)(value / 255.0f), (int)value);
		return txt;
		break;

	default:
		return NULL;
	}
	return NULL;
}

int mi::AllocateTrack(CSequence *pseq, int note)
{
	int c;	

	if ((int)aval.prefertrack != 0) {
		c = (int)(aval.prefertrack - 1);
		if (Tracks[c].CurrentNote == note)
		{
			if (pseq != NULL)
			{
				byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);
				if (*pdata == NOTE_NO)
					return c;
			}
		}
	}

	for (c = 0; c < 12; c++)
	{
		if (Tracks[c].CurrentNote == note)
		{
			if (pseq != NULL)
			{
			byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);
				if (*pdata == NOTE_NO)
					return c;
			}
		}
	}

	int best = -1;

	for (c = 0; c < 12; c++)
	{
		if (Tracks[c].NoteOn == 0)
		{
			if (pseq != NULL)
			{
				byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);
				if (*pdata == NOTE_NO)
					return c;
				
			}

			if (best == -1) {
				best = c;
				return c;
			}
		}
	}

	if (best != -1)
		return best;

	return -1;
}

void mi::MidiNoteOff(int c, CSequence *pseq, int notedelay)
{
	int i2, u, glb;
	Tracks[c].NoteOn = 0;

	if (machine_used1 == 1) {
		glb = pCB->GetMachineInfo(machine_ref1)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref1)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref1)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {
					pCB->ControlChange(machine_ref1, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref1, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used2 == 1) {
		glb = pCB->GetMachineInfo(machine_ref2)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref2)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref2)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref2, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref2, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used3 == 1) {
		glb = pCB->GetMachineInfo(machine_ref3)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref3)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref3)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref3, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref3, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used4 == 1) {
		glb = pCB->GetMachineInfo(machine_ref4)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref4)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref4)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref4, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref4, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used5 == 1) {
		glb = pCB->GetMachineInfo(machine_ref5)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref5)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref5)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref5, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref5, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used6 == 1) {
		glb = pCB->GetMachineInfo(machine_ref6)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref6)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref6)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref6, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref6, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used7 == 1) {
		glb = pCB->GetMachineInfo(machine_ref7)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref7)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref7)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref7, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref7, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used8 == 1) {
		glb = pCB->GetMachineInfo(machine_ref8)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref8)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref8)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref8, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref8, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used9 == 1) {
		glb = pCB->GetMachineInfo(machine_ref9)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref9)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref9)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref9, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref9, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used10 == 1) {
		glb = pCB->GetMachineInfo(machine_ref10)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref10)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref10)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref10, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref10, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used11 == 1) {
		glb = pCB->GetMachineInfo(machine_ref11)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref11)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref11)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref11, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref11, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};
	if (machine_used12 == 1) {
		glb = pCB->GetMachineInfo(machine_ref12)->numGlobalParameters;
		u = (glb + pCB->GetMachineInfo(machine_ref12)->numTrackParameters);
		for (i2 = 0; i2 < u; i2++) {
			if (pCB->GetMachineInfo(machine_ref12)->Parameters[i2]->Type == pt_note) {
				if (i2 >= glb) {	
					pCB->ControlChange(machine_ref12, 2, c, i2 - glb, NOTE_OFF);
				} else {
					pCB->ControlChange(machine_ref12, 1, 0, i2, NOTE_OFF);
				}
			}
		}
	};


	if (pseq != NULL)
	{
		byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);

		if (aval.notedelayused == 1) {
			if (notedelay > 0)
			{
				if (pdata[0] == NOTE_NO) {
					pdata[0] = NOTE_OFF;
					//pdata[2] = 0x0d;
					pdata[2] = (byte)notedelay;
				}
			}
			else
			{
				if (pdata[0] == NOTE_NO) {
					pdata[0] = NOTE_OFF;
					pdata[2] = (byte)0;
				};
			};
		} else {
			if (pdata[0] == NOTE_NO) {
				pdata[0] = NOTE_OFF;
				pdata[2] = (byte)0;
			};
		};
	}
	
}


void mi::MidiNote(int const channel, int const value, int const velocity)
{
	int i2, u, u2, u3, u4, glb;
	int notedelay;

	if (aval.midichannel == channel)
	{
		if (value / 12 > 9) return;	
		bool dontrecord = false;
		
//		int myvelocity;
		CSequence *pseq;

		int reval = value + (aval.miditranspose - 24);
		if (reval < 12) { reval = 12; };
		if (reval > 120) { reval = 120; };
		byte n = (((reval / 12)-1) << 4) | ((reval % 12) + 1);

		if (aval.midirecord == 1) {
			int stateflags = pCB->GetStateFlags();
			
			if (stateflags & SF_PLAYING && stateflags & SF_RECORDING)
				pseq = pCB->GetPlayingSequence(ThisMachine);
			else 
				pseq = NULL;
		} else {
			pseq = NULL;
		};

		if (aval.notedelayused == 1) {
			notedelay = pMasterInfo->PosInTick * 254 / pMasterInfo->SamplesPerTick;
		};

		if (velocity > 0)
		{
			int c = AllocateTrack(pseq, n);

			if (c == -1) {
				c = 0;
				dontrecord = true;
			};

			if (sustaintype == 0) Tracks[c].NoteOn = 1;
			Tracks[c].CurrentNote = n;

			if (machine_used1 == 1) {
				glb = pCB->GetMachineInfo(machine_ref1)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref1)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref1)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {
							pCB->ControlChange(machine_ref1, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref1, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref1)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref1)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref1)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref1, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref1, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used2 == 1) {
				glb = pCB->GetMachineInfo(machine_ref2)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref2)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref2)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref2, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref2, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref2)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref2)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref2)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref2, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref2, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used3 == 1) {
				glb = pCB->GetMachineInfo(machine_ref3)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref3)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref3)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref3, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref3, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref3)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref3)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref3)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref3, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref3, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used4 == 1) {
				glb = pCB->GetMachineInfo(machine_ref4)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref4)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref4)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref4, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref4, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref4)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref4)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref4)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref4, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref4, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used5 == 1) {
				glb = pCB->GetMachineInfo(machine_ref5)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref5)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref5)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref5, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref5, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref5)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref5)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref5)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref5, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref5, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used6 == 1) {
				glb = pCB->GetMachineInfo(machine_ref6)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref6)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref6)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref6, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref6, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref6)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref6)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref6)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref6, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref6, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used7 == 1) {
				glb = pCB->GetMachineInfo(machine_ref7)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref7)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref7)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref7, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref7, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref7)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref7)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref7)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref7, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref7, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used8 == 1) {
				glb = pCB->GetMachineInfo(machine_ref8)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref8)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref8)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref8, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref8, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref8)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref8)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref8)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref8, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref8, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used9 == 1) {
				glb = pCB->GetMachineInfo(machine_ref9)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref9)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref9)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref9, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref9, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref9)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref9)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref9)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref9, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref9, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used10 == 1) {
				glb = pCB->GetMachineInfo(machine_ref10)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref10)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref10)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref10, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref10, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref10)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref10)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref10)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref10, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref10, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used11 == 1) {
				glb = pCB->GetMachineInfo(machine_ref11)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref11)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref11)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref11, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref11, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref11)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref11)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref11)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref11, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref11, 1, 0, i2, u3);
						}
					}
				}
			};
			if (machine_used12 == 1) {
				glb = pCB->GetMachineInfo(machine_ref12)->numGlobalParameters;
				u = (glb + pCB->GetMachineInfo(machine_ref12)->numTrackParameters);
				for (i2 = 0; i2 < u; i2++) {
					if (pCB->GetMachineInfo(machine_ref12)->Parameters[i2]->Type == pt_note) {
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref12, 2, c, i2 - glb, n);
						} else {
							pCB->ControlChange(machine_ref12, 1, 0, i2, n);
						}
					}
				}
				for (i2 = 0; i2 < u; i2++) {
					if (strcmp(pCB->GetMachineInfo(machine_ref12)->Parameters[i2]->Name, vel_name) == 0) {
						u2 = pCB->GetMachineInfo(machine_ref12)->Parameters[i2]->MinValue;
						u4 = 1 + (int)((velocitysens / 2.0f) + ((float)velocity * ((127.0f - velocitysens / 2.0f)/127.0f))); if (u4 > 127) { u4 = 127; };
						u3 = u4 * (pCB->GetMachineInfo(machine_ref12)->Parameters[i2]->MaxValue - u2) / 127 + u2;
						if (i2 >= glb) {	
							pCB->ControlChange(machine_ref12, 2, c, i2 - glb, u3);
						} else {
							pCB->ControlChange(machine_ref12, 1, 0, i2, u3);
						}
					}
				}
			};

			if (aval.midirecord == 1) {
				if (dontrecord == false) {
					if (pseq != NULL)
					{
						byte *pdata = (byte *)pCB->GetPlayingRow(pseq, 2, c);
						pdata[0] = n;
						pdata[1] = velocity;
						if (notedelay > 0)
						{
							if (aval.notedelayused == 1) pdata[2] = (byte)notedelay;
						}
					}
				};
			};
		}
		else
		{
			if (sustaintype == 0) {
				for (int c = 0; c < 12; c++)
				{
					if (Tracks[c].CurrentNote == n)
					{
						MidiNoteOff(c, pseq, notedelay);

						break;
					}	
				}
			}
		}
	}
	
}

#pragma optimize ("", on) 

DLL_EXPORTS

